from base64 import b64encode
from os.path import basename
from typing import List

import click
from click import Path

from common.rpc.mail import send_email


@click.command()
@click.option(
    "--sender",
    default="CS 61A Mailtool",
    help="The sender display name",
)
@click.option(
    "--sender-user",
    default="cs61a",
    help="The sender username (@eecs.berkeley.edu)",
)
@click.option(
    "--target",
    help="The destination email address e.g. cs61a@berkeley.edu",
    prompt=True,
)
@click.option("--subject", help="The subject line of the message", prompt=True)
@click.option("--body", help="The contents of the message", prompt=True)
@click.option(
    "--attachment",
    "attachments",
    type=Path(exists=True, dir_okay=False),
    multiple=True,
    help="Files to attach to the email",
)
def send(
    sender: str,
    sender_user: str,
    target: str,
    subject: str,
    body: str,
    attachments: List[str],
):
    """
    Send emails from eecs.berkeley.edu routed through a university mailserver.
    """
    loaded_attachments = {}

    for attachment in attachments:
        with open(attachment, "rb") as f:
            loaded_attachments[basename(attachment)] = b64encode(f.read()).decode(
                "ascii"
            )

    send_email(
        sender=f"{sender} <{sender_user}@eecs.berkeley.edu>",
        target=target,
        subject=subject,
        body=body,
        attachments=loaded_attachments,
    )



from typing import Dict

from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__, providers=["https://cs162.org/autograder/"])


@requires_master_secret
@service.route("/api/send_email")
def send_email(
    *,
    sender: str,
    target: str,
    subject: str,
    body: str,
    attachments: Dict[str, str] = {},
):
    ...



from typing import Optional

from common.rpc.utils import create_service, requires_access_token

service = create_service(__name__, override="sb")


@requires_access_token
@service.route("/api/update_file")
def update_file(
    *,
    path: str,
    encoded_file_contents: Optional[str] = None,
    symlink: Optional[str] = None,
    delete: bool = False,
):
    ...


@requires_access_token
@service.route("/api/get_server_hashes")
def get_server_hashes():
    ...


@requires_access_token
@service.route("/api/run_make_command", streaming=True)
def run_make_command(*, target: str):
    ...


@requires_access_token
@service.route("/api/is_sandbox_initialized")
def is_sandbox_initialized():
    ...


@requires_access_token
@service.route("/api/initialize_sandbox")
def initialize_sandbox(*, force=True):
    ...



import click, sys, os

from common.rpc.auth_utils import get_token
from common.rpc.buildserver import trigger_build_sync

from github import Github
from github.GithubException import UnknownObjectException

APPS = "Cal-CS-61A-Staff/cs61a-apps"


@click.group("pr")
def pr():
    """
    Commands to interface with PRs.
    """
    pass


@pr.command()
@click.argument("num")
@click.argument("targets", nargs=-1)
def build(num, targets):
    """Build TARGETS for pull request NUM.

    NUM is the PR number you want to build targets
    for, and TARGETS is the list of apps you want
    to build within that PR. If no targets are passed
    in, then all apps modified in the PR are built.
    """
    repo = Github().get_repo(APPS)

    try:
        pull = repo.get_pull(int(num))
    except UnknownObjectException:
        raise Exception("Couldn't find that PR.")

    if pull.state == "closed":
        raise Exception("Cannot build targets for a closed PR!")

    print(f"PR {num}: {pull.title}")
    print(f"Building targets: {targets if targets else 'all'}")

    trigger_build_sync(
        pr_number=int(num),
        target_app=targets,
        _impersonate="buildserver",
        noreply=True,
    )

    print("Build triggered!")



import os
from contextlib import contextmanager
from os import getenv
from typing import List

import sqlalchemy.engine.url

from common.rpc.secrets import get_secret

use_devdb = getenv("ENV", "DEV") in ("DEV", "TEST")
use_prod_proxy = getenv("ENV") == "DEV_ON_PROD"

if use_devdb:
    database_url = "sqlite:///" + os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../app.db"
    )
elif use_prod_proxy:
    curr = os.path.abspath(os.path.curdir)
    prev = None
    while True:
        if ".git" in os.listdir(curr):
            app = os.path.basename(prev)
            break
        prev = curr
        curr = os.path.dirname(curr)
    assert (
        input(
            "Are you sure? You are about to use the PROD database on a DEV build (y/N) "
        )
        .strip()
        .lower()
        == "y"
    )
    if "cloud_sql_proxy" not in os.listdir(curr):
        print(
            "Follow the instructions at https://cloud.google.com/sql/docs/mysql/sql-proxy to install the proxy "
            "in the root directory of the repository, then try again."
        )
        exit(0)
    database_url = sqlalchemy.engine.url.URL(
        drivername="mysql+pymysql",
        username="apps",
        password=get_secret(secret_name="DATABASE_PW"),
        host="127.0.0.1",
        port=3307,
        database=app,
    ).__to_string__(hide_password=False)
else:
    database_url = getenv("DATABASE_URL")

engine = sqlalchemy.create_engine(
    database_url,
    **(
        {}
        if use_devdb
        else dict(pool_size=5, max_overflow=2, pool_timeout=30, pool_recycle=1800)
    )
)


@contextmanager
def connect_db():
    with engine.connect() as conn:

        def db(query: str, args: List[str] = []):
            if use_devdb:
                query = query.replace("%s", "?")
            return conn.execute(query, args)

        yield db


@contextmanager
def transaction_db():
    with engine.begin() as conn:

        def db(*args):
            query, *rest = args
            if use_devdb:
                query = query.replace("%s", "?")
            return conn.execute(query, *rest)

        yield db



from functools import wraps
from typing import List, Union

from common.rpc.secrets import get_secret
from common.rpc.utils import cached, create_service

service = create_service(__name__)


def auth_endpoint(func):
    @wraps(func)
    def wrapped(**kwargs):
        secret = get_secret(secret_name="AUTH_SECRET")
        return func(**kwargs, secret=secret)

    return wrapped


@cached()
@auth_endpoint
@service.route("/admins/is_admin")
def is_admin(*, course: str, email: str, force_course: str = None):
    ...


@cached()
@auth_endpoint
@service.route("/admins/list_admins")
def list_admins(*, course: str):
    ...


@cached()
@service.route("/domains/get_course")
def get_course(*, domain: str):
    ...


@auth_endpoint
@service.route("/google/read_document")
def read_document(*, course: str, url: str, doc_id: str):
    ...


@auth_endpoint
@service.route("/google/read_spreadsheet")
def read_spreadsheet(*, course: str, url: str, doc_id: str, sheet_name: str):
    ...


@auth_endpoint
@service.route("/google/write_spreadsheet")
def write_spreadsheet(
    *,
    course: str,
    url: str,
    doc_id: str,
    sheet_name: str,
    content: List[List[Union[str, float]]]
):
    ...


@cached()
@service.route("/api/list_courses")
def list_courses():
    ...


@cached()
@service.route("/api/get_endpoint")
def get_endpoint(*, course: str):
    ...


@cached()
@service.route("/api/get_endpoint_id")
def get_endpoint_id(*, course: str):
    ...


@cached()
@service.route("/api/validate_secret")
def validate_secret(*, secret: str, course: str):
    ...


@cached()
@auth_endpoint
@service.route("/piazza/perform_action")
def perform_piazza_action(
    *, action: str, course: str, as_staff: bool, is_test: bool, kwargs: dict
):
    ...


@cached()
@auth_endpoint
@service.route("/piazza/course_id")
def piazza_course_id(*, course: str, is_test: bool, test: bool):
    ...


@cached()
@auth_endpoint
@service.route("/slack/workspace_name")
def slack_workspace_name(*, course: str):
    ...


@cached()
@auth_endpoint
@service.route("/slack/post_message")
def post_slack_message(*, course: str, message: str, purpose: str):
    ...


class PiazzaNetwork:
    def __init__(self, course, is_staff, is_test):
        self.course = course
        self.is_staff = is_staff
        self.is_test = is_test

    def __getattr__(self, method):
        def bound_method(**kwargs):
            return perform_piazza_action(
                action=method,
                course=self.course,
                as_staff=self.is_staff,
                is_test=self.is_test,
                kwargs=kwargs,
            )

        return bound_method



import click, requests

from common.rpc.auth_utils import refresh_token, get_token


@click.group("auth")
def auth():
    """
    Commands to interface with authentication.
    """
    pass


@auth.command()
def whoami():
    """
    Get current user information.
    """
    r = requests.get(
        "https://okpy.org/api/v3/user/?access_token={}".format(get_token())
    )
    if not r.ok:
        print("You don't appear to be logged in.")

    data = r.json()["data"]
    print(f"Hi {data.get('name')}! You're logged in as {data.get('email')}.")


@auth.command()
def login():
    """
    Log into Ok and get an access token.
    """
    refresh_token()



import sys
from warnings import warn

import click, os

from sicp.build import build
from sicp.clone import clone
from sicp.venv import venv
from sicp.pr import pr
from sicp.auth import auth
from sicp.send import send

from common.rpc.auth_utils import set_token_path


@click.group()
def cli():
    """
    This is an experimental general-purpose 61A task runner.
    """
    set_token_path(f"{os.path.expanduser('~')}/.sicp_token")


cli.add_command(clone)
cli.add_command(build)
cli.add_command(venv)
cli.add_command(pr)
cli.add_command(auth)
cli.add_command(send)

if sys.version_info[0] == 3 and sys.version_info[1] < 8:
    warn("sicp may not work properly on versions of Python before 3.8")

if __name__ == "__main__":
    cli()



import os
from common.rpc.auth_utils import set_token_path

set_token_path(f"{os.path.expanduser('~')}/.sicp_token")

from common.rpc import auth
from common.rpc import buildserver
from common.rpc import domains
from common.rpc import hosted
from common.rpc import howamidoing
from common.rpc import indexer
from common.rpc import mail
from common.rpc import oh
from common.rpc import paste
from common.rpc import sandbox
from common.rpc import search
from common.rpc import secrets
from common.rpc import sections
from common.rpc import slack



from typing import Union

from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/list_channels")
def list_channels(*, course: str):
    ...


@requires_master_secret
@service.route("/api/post_message")
def post_message(*, course: str, message: Union[str, dict], channel: str):
    ...



import os
import subprocess
import sys
from contextlib import contextmanager
from io import BytesIO
from shutil import rmtree
from typing import List, TextIO, Union
from typing.io import IO


def sh(
    *args,
    env={},
    capture_output=False,
    stream_output=False,
    quiet=False,
    shell=False,
    cwd=None,
    inherit_env=True,
):
    assert not (
        capture_output and stream_output
    ), "Cannot both capture and stream output"

    if shell:
        args = [" ".join(args)]

    if inherit_env:
        env = {**os.environ, **env, "ENV": "dev"}

    if stream_output:
        out = subprocess.Popen(
            args,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            bufsize=1,
            shell=shell,
            cwd=cwd,
        )

        def generator():
            while True:
                line = out.stdout.readline()
                yield line
                returncode = out.poll()
                if returncode is not None:
                    if returncode != 0:
                        # This exception will not be passed to the RPC handler, so we need
                        # to handle it ourselves
                        raise subprocess.CalledProcessError(returncode, args, "", "")
                    else:
                        return ""

        return generator()
    elif capture_output:
        out = subprocess.run(
            args, env=env, capture_output=capture_output, shell=shell, cwd=cwd
        )
    else:
        out = subprocess.run(
            args, env=env, stdout=subprocess.PIPE, shell=shell, cwd=cwd
        )
    if capture_output and not quiet:
        print(out.stdout.decode("utf-8"), file=sys.stdout)
        print(out.stderr.decode("utf-8"), file=sys.stderr)
    out.check_returncode()
    return out.stdout


def clean_all_except(folders: List[str]):
    for dirpath, _, filenames in os.walk("."):
        dirnames = dirpath.split(os.sep)[1:]
        if not dirnames:
            for filename in filenames:
                os.remove(filename)
        elif dirnames[0] not in folders:
            rmtree(dirpath)


@contextmanager
def tmp_directory(*, clean=False):
    main_dir = os.getcwd()
    if clean:
        sh("rm", "-rf", "tmp")
        os.mkdir("tmp")
    try:
        os.chdir("tmp")
        yield None
    finally:
        os.chdir(main_dir)


@contextmanager
def redirect_descriptor(
    src: Union[IO, TextIO, BytesIO], target: Union[IO, TextIO, BytesIO]
):
    src_fd: int = src.fileno()
    with os.fdopen(os.dup(src_fd), "wb") as saved:
        src.flush()
        os.dup2(target.fileno(), src_fd)
        try:
            yield src
        finally:
            src.flush()
            os.dup2(saved.fileno(), src_fd)



import json
import traceback
from functools import wraps

import flask
import requests
from cachetools import TTLCache
from flask import Response, has_request_context, jsonify, request, stream_with_context

from common.rpc.auth_utils import get_token, refresh_token
from common.secrets import get_master_secret

STATUS_MARKER = "__INTERNAL_STATUS_MARKER"


class Service:
    def __init__(self, route):
        self.route = route


def create_service(app: str, override=None, providers=None):
    app = override or app.split(".")[-1]

    def route(path, *, streaming=False):
        def decorator(func):
            @wraps(func)
            def wrapped(**kwargs):
                noreply = kwargs.pop("noreply", False)

                if providers:
                    endpoints = [f"{provider}{path}" for provider in providers]
                else:
                    endpoints = []
                    if has_request_context() and not noreply:
                        proxied_host = request.headers.get("X-Forwarded-For-Host")
                        if proxied_host:
                            parts = proxied_host.split(".")
                            if "pr" in parts:
                                pr = parts[0]
                                endpoints.append(
                                    f"https://{pr}.{app}.pr.cs61a.org{path}"
                                )
                    endpoints.append(f"https://{app}.cs61a.org{path}")

                if (
                    not get_master_secret()
                    and "_impersonate" in kwargs
                    and not "_sudo_token" in kwargs
                ):
                    from common.rpc.secrets import (
                        get_secret_from_server,
                    )  # placed here to avoid circular imports

                    print(f"Attempting to impersonate {kwargs.get('_impersonate')}")

                    try:
                        sudo_secret = get_secret_from_server(
                            secret_name="MASTER",
                            _impersonate=kwargs.pop("_impersonate"),
                            _sudo_token=get_token(),
                        )
                    except PermissionError:
                        refresh_token()
                        try:  # second attempt, in case the first was just an expired token
                            sudo_secret = get_secret_from_server(
                                secret_name="MASTER",
                                _impersonate=kwargs.pop("_impersonate"),
                                _sudo_token=get_token(),
                            )
                        except PermissionError:
                            raise PermissionError(
                                "You must be logged in as an admin to do that."
                            )

                    kwargs["master_secret"] = sudo_secret

                for i, endpoint in enumerate(endpoints):
                    if noreply:
                        try:
                            requests.post(endpoint, json=kwargs, timeout=1)
                        except requests.exceptions.ReadTimeout:
                            pass
                    else:
                        try:
                            resp = requests.post(
                                endpoint, json=kwargs, stream=streaming
                            )
                            if i != len(endpoints) - 1:
                                # if a PR build reports failure, try the prod build
                                resp.raise_for_status()
                        except:
                            if i != len(endpoints) - 1:
                                # on a PR build, try the main endpoint next
                                continue
                            else:
                                raise
                        if resp.status_code == 401:
                            raise PermissionError(resp.text)
                        elif resp.status_code == 500:
                            raise Exception(resp.text)
                        resp.raise_for_status()
                        if streaming:

                            def generator():
                                buffer = []
                                ok = True
                                for x in resp.iter_content():
                                    buffer.append(x.decode("ascii"))
                                    buff_string = "".join(buffer)
                                    if STATUS_MARKER in buff_string:
                                        # We are now reading in the error message
                                        # Stop flushing the buffer
                                        ok = False
                                    if ok and len(buff_string) > len(STATUS_MARKER):
                                        yield buff_string[: -len(STATUS_MARKER)]
                                        buffer = [buff_string[-len(STATUS_MARKER) :]]
                                buff_string = "".join(buffer)
                                if not buff_string.endswith(STATUS_MARKER):
                                    # some error occurred
                                    pos = buff_string.index(STATUS_MARKER)
                                    raise Exception(
                                        buff_string[pos + len(STATUS_MARKER) :]
                                    )
                                yield from buffer[: -len(STATUS_MARKER)]

                            return generator()
                        else:
                            return resp.json()

            def bind(app: flask.Flask):
                def decorator(func):
                    def handler():
                        kwargs = request.json
                        try:
                            out = func(**kwargs)
                            if streaming:
                                # we should stream our response
                                def generator():
                                    try:
                                        for x in out:
                                            yield bytes(
                                                x, encoding="ascii", errors="replace"
                                            )
                                    except Exception as e:
                                        yield STATUS_MARKER
                                        yield bytes(
                                            str(e), encoding="ascii", errors="replace"
                                        )
                                    else:
                                        yield STATUS_MARKER

                                return Response(stream_with_context(generator()))
                            else:
                                return jsonify(out)
                        except PermissionError as e:
                            return str(e), 401
                        except Exception as e:
                            traceback.print_exc()
                            return str(e), 500

                    app.add_url_rule(path, func.__name__, handler, methods=["POST"])
                    return func

                return decorator

            wrapped.bind = bind

            return wrapped

        return decorator

    return Service(route)


def requires_master_secret(func):
    @wraps(func)
    def wrapped(**kwargs):
        return func(**kwargs, master_secret=get_master_secret())

    return wrapped


def requires_access_token(func):
    @wraps(func)
    def wrapped(**kwargs):
        try:
            return func(**kwargs, access_token=get_token())
        except PermissionError:
            refresh_token()
            return func(**kwargs, access_token=get_token())

    return wrapped


def cached(ttl: int = 1800):
    """
    Caches the return value of this RPC method for `ttl` seconds (defaults to 1800s)
    """
    cache = TTLCache(1000, ttl)

    def decorator(func):
        @wraps(func)
        def wrapped(**kwargs):
            key = json.dumps(kwargs)
            if key not in cache:
                cache[key] = func(**kwargs)
            return cache[key]

        return wrapped

    return decorator



import re, requests
from os import getenv

from cachetools import TTLCache
from flask import request

import common.rpc as rpc

DOMAIN_COURSES = TTLCache(1000, 1800)
COURSE_ENDPOINTS = TTLCache(1000, 1800)
ENDPOINT_ID = TTLCache(1000, 1800)


def get_course(domain=None):
    if getenv("ENV") != "prod":
        return "cs61a"
    if not domain:
        domain = get_domain()
    if "pr" in domain:
        DOMAIN_COURSES[domain] = "cs61a"
    if domain not in DOMAIN_COURSES:
        DOMAIN_COURSES[domain] = rpc.auth.get_course(domain=domain)
    return DOMAIN_COURSES[domain]


def get_domain():
    return request.headers.get("X-Forwarded-For-Host") or request.headers["HOST"]


def get_endpoint(course=None):
    if getenv("ENV") != "prod":
        return "cal/cs61a/sp21"
    if not course:
        course = get_course()
    if course not in COURSE_ENDPOINTS:
        COURSE_ENDPOINTS[course] = rpc.auth.get_endpoint(course=course)
    return COURSE_ENDPOINTS[course]


def get_course_id(course=None):
    if getenv("ENV") != "prod":
        return 1
    if not course:
        course = get_course()
    if course not in ENDPOINT_ID:
        ENDPOINT_ID[course] = rpc.auth.get_endpoint_id(course=course)
    return ENDPOINT_ID[course]


def is_admin(email, course=None):
    if getenv("ENV") != "prod":
        return True
    if not course:
        course = get_course()
    return rpc.auth.is_admin(email=email, course=course, force_course=course)


def is_admin_token(access_token, course=None):
    ret = requests.get(
        "https://okpy.org/api/v3/user/", params={"access_token": access_token}
    )
    return ret.status_code == 200 and is_admin(
        ret.json()["data"]["email"],
        course=course,
    )


def format_coursecode(course):
    m = re.match(r"([a-z]+)([0-9]+[a-z]?)", course)
    return m and (m.group(1) + " " + m.group(2)).upper()



from typing import Dict, Optional

from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__, "deploy.hosted")


@requires_master_secret
@service.route("/api/list_apps")
def list_apps():
    ...


@requires_master_secret
@service.route("/api/new")
def new(*, img: str, name: Optional[str] = None, env: Dict[str, str] = {}):
    ...


@requires_master_secret
@service.route("/api/stop")
def stop(*, name: str):
    ...


@requires_master_secret
@service.route("/api/run")
def run(*, name: str):
    ...


@requires_master_secret
@service.route("/api/delete")
def delete(*, name: str):
    ...


@requires_master_secret
@service.route("/api/add_domain")
def add_domain(
    *, name: str, domain: str, force: bool = False, proxy_set_header: dict = {}
):
    ...


@requires_master_secret
@service.route("/api/service_log")
def service_log():
    ...


@requires_master_secret
@service.route("/api/container_log")
def container_log(*, name: str):
    ...


@requires_master_secret
@service.route("/api/create_pr_subdomain")
def create_pr_subdomain(*, app: str, pr_number: int, pr_host: str):
    ...



from flask import current_app


def make_row(content, target, action="Remove"):
    return f"""<form style="display: inline" action="{target}" method="post">
            {content}
            <input type="submit" value="{action}">
    </form>"""


def html(out):
    # trivial changes to make things look less ugly
    if "<h1>" not in out:
        if hasattr(current_app, "remote"):
            header = current_app.remote.consumer_key
            if header.startswith("61a-"):
                header = header[3:]
            header = header.replace("-", " ").title()
            out = f"<h1>61A {header}</h1>" + out
    return f"""
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/spcss@0.5.0">
{out}
"""


def error(out):
    report = f"<pre>{out}</pre>" if out else ""
    return html(
        f"<h2 style='color: red'>Something went wrong.</h2>{report} "
        f'If you are on 61A staff, visit <a href="https://logs.cs61a.org">logs.cs61a.org</a> '
        f"to see crash logs. Otherwise, please post in the EECS Crossroads Slack for help."
    )



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()


setup(
    name="sicp",
    version="0.1.11",
    author="Rahul Arya, Vanshaj Singhania",
    author_email="rahularya@berkeley.edu",
    long_description=readme,
    long_description_content_type="text/markdown",
    licence="MIT",
    packages=find_packages(include=["sicp", "sicp.common", "sicp.common.rpc"]),
    package_data={"": ["**/*.tex"]},
    include_package_data=True,
    entry_points={"console_scripts": ["sicp=sicp.__main__:cli"]},
    python_requires=">=3.6",
    install_requires=[
        "click",
        "watchdog",
        "crcmod",
        "tqdm",
        "flask",
        "cachetools",
        "colorama",
        "requests",
        "PyGithub",
    ],
)



from os import getenv
from urllib.parse import urlparse

from flask import request, url_for as flask_url_for


def get_host() -> str:
    return request.headers.get("X-Forwarded-For-Host") or request.headers.get("Host")


def url_for(*args, **kwargs) -> str:
    host = get_host()
    redirect_url = urlparse(flask_url_for(*args, **kwargs))
    # noinspection PyProtectedMember
    return redirect_url._replace(
        netloc=host, scheme="https" if getenv("ENV") == "prod" else "http"
    ).geturl()
